// TVTuner.c
//
// Requires DDK Only
// File created on 8/12/2008
//

#include "pch.h"
#include "..\intrface.h"
#ifdef TVTUNER_WMI_TRACE
#include "TVTuner.tmh"
#endif

// global data
TVTUNER_DATA g_Data;

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DriverEntry 
//      Installable driver initialization entry point.
//      This entry point is called directly by the I/O system.
//
//  Arguments:
//      IN  DriverObject
//              pointer to the driver object
//
//      IN  RegistryPath
//              pointer to a unicode string representing the path,
//              to driver-specific key in the registry.
//
//  Return Value:
//      Status
//
NTSTATUS DriverEntry(
    IN  PDRIVER_OBJECT  DriverObject,
    IN  PUNICODE_STRING RegistryPath
    )
{
    NTSTATUS    status;

    TVTunerDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"++");
    TVTunerDebugPrint(DBG_INIT, DBG_INFO, "Compiled at %s on %s", __TIME__, __DATE__);

#ifdef DBG
//    DbgBreakPoint();
#endif

#ifdef TVTUNER_WMI_TRACE 
    WPP_INIT_TRACING(DriverObject, RegistryPath);
#endif

    status = STATUS_SUCCESS;

    RtlZeroMemory(&g_Data, sizeof(TVTUNER_DATA));

    // save registry path for wmi
    g_Data.RegistryPath.Length = RegistryPath->Length;
    g_Data.RegistryPath.MaximumLength = RegistryPath->Length + sizeof(UNICODE_NULL);
    g_Data.RegistryPath.Buffer = (PWCHAR)ExAllocatePoolWithTag(
                                            PagedPool,
                                            g_Data.RegistryPath.MaximumLength,
                                            TVTUNER_POOL_TAG
                                            );

    if (g_Data.RegistryPath.Buffer == NULL)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

        TVTunerDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__": Failed to allocate memory for RegistryPath");

        return status;
    }

    RtlCopyUnicodeString(&g_Data.RegistryPath, RegistryPath);

    // detect current operating system
    if (IoIsWdmVersionAvailable(1, 0x30))
    {
        g_Data.WdmVersion = 0x0130;
    }
    else if (IoIsWdmVersionAvailable(1, 0x20))
    {
        g_Data.WdmVersion = 0x0120;
    }
    else if (IoIsWdmVersionAvailable(1, 0x10))
    {
        g_Data.WdmVersion = 0x0110;
    }
    else if (IoIsWdmVersionAvailable(1, 0x05))
    {
        g_Data.WdmVersion = 0x0105;
    }
    else
    {
        g_Data.WdmVersion = 0x0100;
    }

    // setup our dispatch function table in the driver object
    DriverObject->MajorFunction[IRP_MJ_PNP] = TVTunerPnpDispatch;
    DriverObject->MajorFunction[IRP_MJ_POWER] = TVTunerPowerDispatch;
    DriverObject->MajorFunction[IRP_MJ_CREATE] = TVTunerCreateDispatch;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = TVTunerCloseDispatch;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = TVTunerDeviceIoControlDispatch;
    DriverObject->MajorFunction[IRP_MJ_READ] = TVTunerReadDispatch;
    DriverObject->MajorFunction[IRP_MJ_WRITE] = TVTunerWriteDispatch;
    DriverObject->MajorFunction[IRP_MJ_CLEANUP] = TVTunerCleanupDispatch;
    DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = TVTunerSystemControlDispatch;

    DriverObject->DriverExtension->AddDevice = TVTunerAddDevice;
    DriverObject->DriverUnload = TVTunerUnload;

    TVTunerDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerAddDevice 
//      The PnP manager new device callback
//
//  Arguments:
//      IN  DeviceObject
//              pointer to a device object.
//
//      IN  PhysicalDeviceObject
//              pointer to a device object created by the
//              underlying bus driver.
//
//  Return Value:
//      Status
//
NTSTATUS TVTunerAddDevice(
    IN PDRIVER_OBJECT   DriverObject,
    IN PDEVICE_OBJECT   PhysicalDeviceObject
    )
{
    NTSTATUS                            status;
    PDEVICE_OBJECT                      deviceObject;
    PTVTUNER_DEVICE_EXTENSION			deviceExtension;
    POWER_STATE                         powerState;

    TVTunerDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"++: PDO %p", PhysicalDeviceObject);

    // Create our function device object.
    status = IoCreateDevice(
                DriverObject,
                sizeof (TVTUNER_DEVICE_EXTENSION),
                NULL,                    // no name
                FILE_DEVICE_UNKNOWN,
                FILE_DEVICE_SECURE_OPEN, // Do not use if binary intended for 9x
                FALSE,
                &deviceObject
                );

    if (!NT_SUCCESS(status)) 
    {
        TVTunerDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. IoCreateDevice STATUS %x", status);

        return status;
    }

    TVTunerDebugPrint(DBG_INIT, DBG_INFO, __FUNCTION__": Created FDO %p", deviceObject);

    // Initialize the device extension.
    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)deviceObject->DeviceExtension;

    // Zero the memory
    RtlZeroMemory(deviceExtension, sizeof(TVTUNER_DEVICE_EXTENSION));

    // save the PDO pointer
    deviceExtension->PhysicalDeviceObject = PhysicalDeviceObject;

    // save our device object
    deviceExtension->DeviceObject = deviceObject;

    // Initialize the powering down flag
    deviceExtension->bPowerStop = FALSE;

    // set RemoveCount to 1. Transition to zero
    // means IRP_MN_REMOVE_DEVICE was received
    deviceExtension->RemoveCount = 1;

    // Initialize Remove event
    KeInitializeEvent(&deviceExtension->RemoveEvent, NotificationEvent, FALSE);

    // initialize io lock
    TVTunerInitializeIoLock(&deviceExtension->IoLock, deviceObject);

    deviceExtension->PnpState = PnpStateNotStarted;
    deviceExtension->PreviousPnpState = PnpStateNotStarted;

    // allocate WorkItem for power processing
    deviceExtension->PowerWorkItem = IoAllocateWorkItem(deviceExtension->DeviceObject);
    if (deviceExtension->PowerWorkItem == NULL)
    {
        TVTunerDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. Failed to allocated PowerWorkItem");

        IoDeleteDevice(deviceObject);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    // Initialize the device object flags

    // All WDM drivers are supposed to set this flag.  Devices
    // requiring a large amount of current at statup set the 
    // DO_POWER_INRUSH flag instead.  These flags are mutually
    // exclusive.
    if (PhysicalDeviceObject->Flags & DO_POWER_PAGABLE) 
    {
        deviceObject->Flags |= DO_POWER_PAGABLE;
    }

    // This flag sets the buffering method for reads and writes
    // to METHOD_DIRECT.  IOCTLs are handled by IO control codes
    // independent of the value of this flag.
    deviceObject->Flags |= DO_DIRECT_IO;

    // Typically, (almost always) the function driver for a device is the
    // power policy owner for that device.  In some cases
    // another driver or system component may assume this role.

    // Set the initial device power state of the device to powered down.
    deviceExtension->DevicePowerState = PowerDeviceD3;
    powerState.DeviceState = deviceExtension->DevicePowerState;

    // the initial system power state must be working or we wouldn't be here.
    deviceExtension->SystemPowerState = PowerSystemWorking;

    // Call PoSetPowerState to notify the power manager of our initial power state.
    PoSetPowerState(deviceObject, DevicePowerState, powerState);

    // Attach our device to the device stack and get the device object to 
    // which we send down IRPs to.
    deviceExtension->LowerDeviceObject = IoAttachDeviceToDeviceStack(
                                            deviceObject,
                                            PhysicalDeviceObject
                                            );

    if (deviceExtension->LowerDeviceObject == NULL) 
    {
        TVTunerDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. IoAttachDeviceToDeviceStack failed");

        IoFreeWorkItem(deviceExtension->PowerWorkItem);
        IoDeleteDevice(deviceObject);
        return STATUS_DEVICE_REMOVED;
    }

    // Initialize our device's interface
    status = IoRegisterDeviceInterface (
                PhysicalDeviceObject,
                (LPGUID)&GUID_DEVINTERFACE_TVTUNER,
                NULL,
                &deviceExtension->InterfaceName
                );

    if (!NT_SUCCESS(status)) 
    {
        TVTunerDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. IoRegisterDeviceInterface STATUS %x", status);

        IoDetachDevice(deviceExtension->LowerDeviceObject);
        IoFreeWorkItem(deviceExtension->PowerWorkItem);
        IoDeleteDevice(deviceObject);
        return status;
    }

    // increment device instance count
    ++g_Data.InstanceCount;

    // We are all done, so we need to clear the
    // DO_DEVICE_INITIALIZING flag.  This must be our
    // last action in AddDevice
    deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

    TVTunerDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"--. TVTunerAddDevice--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerStartDevice
//      Start device handler, sets up resources
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  Irp
//              pointer to the Start IRP
//
//  Return Value:
//      NT status code
//
NTSTATUS TVTunerStartDevice(
    IN  PTVTUNER_DEVICE_EXTENSION    DeviceExtension,
    IN  PIRP                            Irp
    )
{
    NTSTATUS                        status;
    PURB                            urb;
    PUSB_DEVICE_DESCRIPTOR          deviceDescriptor;
    PUSB_CONFIGURATION_DESCRIPTOR   configDescriptor;
    USHORT                          size;
    PUSBD_INTERFACE_LIST_ENTRY      interfaceBuffer;
    PUSBD_INTERFACE_LIST_ENTRY      interfaceList;
    UCHAR                           numInterfaces;
    UCHAR                           index;
    PUSB_INTERFACE_DESCRIPTOR       interfaceDescriptor;
    PUSBD_INTERFACE_INFORMATION     interfaceInfo;
    PUSB_ENDPOINT_DESCRIPTOR        epDescriptor;

    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    // we will use do{}while(FALSE); structure
    // to help with resource cleanup
    urb = NULL;
    deviceDescriptor = NULL;
    configDescriptor = NULL;
    interfaceBuffer = NULL;

    do
    {
        // we need to allocate an URB to talk
        // to usb bus driver
        urb = (PURB)ExAllocatePoolWithTag(
                        NonPagedPool,
                        sizeof(URB),
                        TVTUNER_POOL_TAG
                        );


        TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" sizeof(URB) is  %d", sizeof(URB));


        if (urb == NULL)
        {
	    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" STATUS_INSUFFICIENT_RESOURCES");

            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        // first let's read device descriptor
        deviceDescriptor = (PUSB_DEVICE_DESCRIPTOR)ExAllocatePoolWithTag(
                                                    NonPagedPool,
                                                    sizeof(USB_DEVICE_DESCRIPTOR),
                                                    TVTUNER_POOL_TAG
                                                    );

        TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" sizeof(USB_DEVICE_DESCRIPTOR) is  %d", sizeof(USB_DEVICE_DESCRIPTOR));

        if (deviceDescriptor == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
		    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" STATUS_INSUFFICIENT_RESOURCES");

            break;
        }

        TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST) is  %d", sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST));

		

        UsbBuildGetDescriptorRequest(
            urb,
            (USHORT) sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
            USB_DEVICE_DESCRIPTOR_TYPE,
            0,
            0,
            deviceDescriptor,
            NULL,
            sizeof(USB_DEVICE_DESCRIPTOR),
            NULL
            );

        status = TVTunerSubmitUrbSynch(DeviceExtension, urb);

        if (!NT_SUCCESS(status))
        {
			TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" SubmitUrbSync is failed 1");
            break;
        } 


        // We have to read configuration descriptor twice.
        // First time to obtain its full length and second time
        // to read full configuration descriptor
        configDescriptor = (PUSB_CONFIGURATION_DESCRIPTOR)ExAllocatePoolWithTag(
                                                            NonPagedPool,
                                                            sizeof(USB_CONFIGURATION_DESCRIPTOR),
                                                            TVTUNER_POOL_TAG
                                                            );

        if (configDescriptor == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
	    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" STATUS_INSUFFICIENT_RESOURCES");

            break;
        }

        UsbBuildGetDescriptorRequest(
            urb,
            (USHORT)sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
            USB_CONFIGURATION_DESCRIPTOR_TYPE,
            0,
            0,
            configDescriptor,
            NULL,
            sizeof(USB_CONFIGURATION_DESCRIPTOR),
            NULL
            );

        status = TVTunerSubmitUrbSynch(DeviceExtension, urb);
        if (!NT_SUCCESS(status))
        {
	    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" TVTunerSubmitUrbSynch is failed 2");
            break;
        }

        size = configDescriptor->wTotalLength;
        ExFreePool(configDescriptor);

        configDescriptor = (PUSB_CONFIGURATION_DESCRIPTOR)ExAllocatePoolWithTag(
                                                            NonPagedPool,
                                                            size,
                                                            TVTUNER_POOL_TAG
                                                            );

        if (configDescriptor == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
	    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" STATUS_INSUFFICIENT_RESOURCES");

            break;
        }

        UsbBuildGetDescriptorRequest(
            urb,
            (USHORT)sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
            USB_CONFIGURATION_DESCRIPTOR_TYPE,
            0,
            0,
            configDescriptor,
            NULL,
            size,
            NULL
            );

        status = TVTunerSubmitUrbSynch(DeviceExtension, urb);
        if (!NT_SUCCESS(status))
        {
	    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" TVTunerSubmitUrbSynch is failed 3");

            break;
        }

        // now that we have configuration descriptor lets find our interface
        // and enable it
        numInterfaces = configDescriptor->bNumInterfaces;
        interfaceBuffer = interfaceList =
            (PUSBD_INTERFACE_LIST_ENTRY)ExAllocatePoolWithTag(
                                            NonPagedPool,
                                            sizeof(USBD_INTERFACE_LIST_ENTRY) * (numInterfaces + 1),
                                            TVTUNER_POOL_TAG
                                            );

        if (interfaceBuffer == NULL)
        {

	    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" STATUS_INSUFFICIENT_RESOURCES");

            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        for (index = 0; index < numInterfaces; ++index)
        {
            interfaceDescriptor = USBD_ParseConfigurationDescriptorEx(
                                    configDescriptor,
                                    configDescriptor,
                                    index,
                                    0,
                                    -1,
                                    -1,
                                    -1
                                    );

            if (interfaceDescriptor != NULL)
            {
                interfaceList->InterfaceDescriptor = interfaceDescriptor;
                interfaceList->Interface = NULL;

                ++interfaceList;
            }
        }

        interfaceList->InterfaceDescriptor = NULL;
        interfaceList->Interface = NULL;

        ExFreePool(urb);
        urb = USBD_CreateConfigurationRequestEx(configDescriptor, interfaceBuffer);
        if (urb == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        interfaceInfo = &urb->UrbSelectConfiguration.Interface;

        // verify configuration
        interfaceDescriptor = interfaceBuffer->InterfaceDescriptor;
        if (interfaceDescriptor->bNumEndpoints != 2)
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
            break;
        }

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)interfaceDescriptor;

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)USBD_ParseDescriptors(
                                                    configDescriptor,
                                                    configDescriptor->wTotalLength,
                                                    epDescriptor,
                                                    USB_ENDPOINT_DESCRIPTOR_TYPE
                                                    );

        if ((epDescriptor == NULL) ||
            (epDescriptor->bEndpointAddress != 0x81) ||
            (epDescriptor->bmAttributes != USB_ENDPOINT_TYPE_BULK) ||
            (epDescriptor->wMaxPacketSize != 0x200))
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;

	    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" STATUS_DEVICE_CONFIGURATION_ERROR");

            break;
        }
        else
        {
            interfaceInfo->Pipes[0].MaximumTransferSize = 4096;
            ++epDescriptor;
        }

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)USBD_ParseDescriptors(
                                                    configDescriptor,
                                                    configDescriptor->wTotalLength,
                                                    epDescriptor,
                                                    USB_ENDPOINT_DESCRIPTOR_TYPE
                                                    );

        if ((epDescriptor == NULL) ||
            (epDescriptor->bEndpointAddress != 0x1) ||
            (epDescriptor->bmAttributes != USB_ENDPOINT_TYPE_BULK) ||
            (epDescriptor->wMaxPacketSize != 0x200))
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
	    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" STATUS_DEVICE_CONFIGURATION_ERROR");

            break;
        }
        else
        {
            interfaceInfo->Pipes[1].MaximumTransferSize = 4096;
            ++epDescriptor;
        }

        status = TVTunerSubmitUrbSynch(DeviceExtension, urb);
        if (!NT_SUCCESS(status))
        {
            break;
        }

        DeviceExtension->InterfaceInformation =
            (PUSBD_INTERFACE_INFORMATION)ExAllocatePoolWithTag(
                                            NonPagedPool,
                                            interfaceInfo->Length,
                                            TVTUNER_POOL_TAG
                                            );

        if (DeviceExtension->InterfaceInformation == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        RtlCopyMemory(DeviceExtension->InterfaceInformation, interfaceInfo, interfaceInfo->Length);

        DeviceExtension->ConfigHandle = urb->UrbSelectConfiguration.ConfigurationHandle;
        DeviceExtension->ConfigDescriptor = configDescriptor;
        configDescriptor = NULL;
    }
    while (FALSE);

    if (urb != NULL)
    {
        ExFreePool(urb);
    }

    if (deviceDescriptor != NULL)
    {
        ExFreePool(deviceDescriptor);
    }

    if (configDescriptor != NULL)
    {
        ExFreePool(configDescriptor);
    }

    if (interfaceBuffer != NULL)
    {
        ExFreePool(interfaceBuffer);
    }

    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerFreeResources
//      This routine returns all the USB resources acquired during
//      device startup.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      NT status code
//
NTSTATUS TVTunerFreeResources(
    IN  PTVTUNER_DEVICE_EXTENSION DeviceExtension
    )
{
    NTSTATUS    status = STATUS_SUCCESS;
    PURB        urb;

    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"++");
    if (DeviceExtension->ConfigDescriptor != NULL)
    {
        if (DeviceExtension->PnpState != PnpStateSurpriseRemoved)
        {
            urb = (PURB)ExAllocatePoolWithTag(
                            NonPagedPool,
                            sizeof(struct _URB_SELECT_CONFIGURATION),
                            TVTUNER_POOL_TAG
                            );

            if (urb != NULL)
            {
                UsbBuildSelectConfigurationRequest(
                    urb,
                    (USHORT)sizeof(struct _URB_SELECT_CONFIGURATION),
                    NULL
                    );

                status = TVTunerSubmitUrbSynch(DeviceExtension, urb);

                ExFreePool(urb);
            }
            else
            {
                status = STATUS_INSUFFICIENT_RESOURCES;
            }
        }

        if (DeviceExtension->ConfigDescriptor != NULL)
        {
            ExFreePool(DeviceExtension->ConfigDescriptor);
            DeviceExtension->ConfigDescriptor = NULL;
        }
    }

    if (DeviceExtension->InterfaceInformation != NULL)
    {
        ExFreePool(DeviceExtension->InterfaceInformation);
        DeviceExtension->InterfaceInformation = NULL;
    }

    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"--");

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerCreateDispatch
//      Dispatch routine for IRP_MJ_CREATE requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the create IRP
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerCreateDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    NTSTATUS                status;

    TVTunerDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__": IRP %p", Irp);

    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    if (!TVTunerAcquireRemoveLock(deviceExtension))
    {
        status = STATUS_NO_SUCH_DEVICE;

        Irp->IoStatus.Status = status;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);

        TVTunerDebugPrint(DBG_CREATECLOSE, DBG_WARN, __FUNCTION__"$$--. IRP %p, STATUS %x", Irp, status);

        return status;
    }

    // increment open count
    InterlockedIncrement(&deviceExtension->OpenHandleCount);

    status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    TVTunerReleaseRemoveLock(deviceExtension);

    TVTunerDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"--. IRP %p, STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerCloseDispatch
//      Dispatch routine for IRP_MJ_CLOSE requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the close IRP
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerCloseDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    NTSTATUS                status;

    TVTunerDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = status;
    IoCompleteRequest (Irp, IO_NO_INCREMENT);

    InterlockedDecrement(&deviceExtension->OpenHandleCount);

    TVTunerDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"--. IRP %p, STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerCleanupDispatch
//      Dispatch routine for IRP_MJ_CLEANUP requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the cleanup IRP
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerCleanupDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    NTSTATUS                status;
    PIO_STACK_LOCATION      irpStack;

    TVTunerDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    irpStack = IoGetCurrentIrpStackLocation(Irp);
    TVTunerFlushQueues(deviceExtension, irpStack->FileObject);

    status = STATUS_SUCCESS;

    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    TVTunerDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerUnload
//      Driver Unload routine.
//
//  Arguments:
//      IN  DriverObject
//              pointer to the driver object
//
//  Return Value:
//      none
//
VOID TVTunerUnload(
    IN  PDRIVER_OBJECT  DriverObject
    )
{
    TVTunerDebugPrint(DBG_UNLOAD, DBG_TRACE, __FUNCTION__"++");

    // The device object(s) should be NULL now
    // (since we unload, all the devices objects associated with this
    // driver must be deleted.
    ASSERT(DriverObject->DeviceObject == NULL);

    // We should not be unloaded until all the devices we control
    // have been removed from our queue.

    // release memory block allocated for registry path
    if (g_Data.RegistryPath.Buffer != NULL)
    {
        ExFreePool(g_Data.RegistryPath.Buffer);
        g_Data.RegistryPath.Buffer = NULL;
    }

    TVTunerDebugPrint(DBG_UNLOAD, DBG_TRACE, __FUNCTION__"--");

#ifdef TVTUNER_WMI_TRACE
    WPP_CLEANUP(DriverObject);
#endif

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerIsStoppable
//      This routine determines whether the device can be safely stopped. 
//      In our particular case, we'll assume we can always stop the device.
//      A device might fail the request if it doesn't have a queue for the
//      requests that might come or if it was notified that it is in the 
//      paging path.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      returns TRUE if the device is stoppable, FALSE otherwise
//
BOOLEAN TVTunerIsStoppable(
    IN  PTVTUNER_DEVICE_EXTENSION    DeviceExtension
    )
{
    return TRUE;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerIsRemovable
//      This routine determines whether the device can be safely removed. 
//      A device shouldn't be removed if, for example, it has open handles or
//      removing the device could result in losing data (plus the reasons
//      mentioned in the TVTunerIsStoppable function comments). The PnP manager on 
//      Windows 2000 fails on its own any attempt to remove, if there any 
//      open handles to the device.  However on Win9x, the driver must keep 
//      count of open handles and fail QueryRemove if there are any open 
//      handles.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      returns TRUE if the device is stoppable, FALSE otherwise
//
BOOLEAN TVTunerIsRemovable(
    IN  PTVTUNER_DEVICE_EXTENSION    DeviceExtension
    )
{
    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerSubmitIrpSyncComplete
//      Completion routine for sync IRP requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to our device object
//
//      IN  Irp
//              pointer to the PnP IRP
//
//      IN  Context
//              our event used to signal IRP completion
//
//  Return Value:
//      STATUS_MORE_PROCESSING_REQUIRED
//
NTSTATUS  TVTunerSubmitIrpSyncComplete(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp,
    IN  PVOID           Context
    )
{
    PKEVENT event = (PKEVENT)Context;

    // If the lower driver didn't return STATUS_PENDING, we don't need to 
    // set the event because we won't be waiting on it. 
    if (Irp->PendingReturned) 
    {
        KeSetEvent(event, IO_NO_INCREMENT, FALSE);
    }

    return STATUS_MORE_PROCESSING_REQUIRED;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerSubmitIrpSync
//      Sends the given IRP down the stack to the next lower driver and 
//      waits in a synchronous fashion for the IRP to complete
//
//  Arguments:
//      IN  DeviceObject
//              Pointer to device object for our device
//
//      IN  Irp
//              IRP to send down
//
//  Return Value:
//      NT status code
//
NTSTATUS TVTunerSubmitIrpSync(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    KEVENT   event;
    NTSTATUS status;

    KeInitializeEvent(&event, NotificationEvent, FALSE);

    IoCopyCurrentIrpStackLocationToNext(Irp);

    IoSetCompletionRoutine(
        Irp,
        TVTunerSubmitIrpSyncComplete,
        &event,
        TRUE,
        TRUE,
        TRUE
        );

    status = IoCallDriver(DeviceObject, Irp);

    // Wait for lower drivers to be done with the Irp.
    // Important thing to note here is when you allocate
    // memory for an event in the stack you must do a
    // KernelMode wait instead of UserMode to prevent
    // the stack from getting paged out.

    if (status == STATUS_PENDING) 
    {
       KeWaitForSingleObject(
           &event,
           Executive,
           KernelMode,
           FALSE,
           NULL
           );

       status = Irp->IoStatus.Status;
    }

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerAcquireRemoveLock
//      Acquires remove lock.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      FALSE if remove device pending, TRUE otherwise.
//
BOOLEAN TVTunerAcquireRemoveLock(
    IN  PTVTUNER_DEVICE_EXTENSION    DeviceExtension
    )
{
    LONG    count;

    count = InterlockedIncrement(&DeviceExtension->RemoveCount);

    ASSERT(count > 0);

    if (DeviceExtension->PnpState == PnpStateRemoved)
    {
        TVTunerReleaseRemoveLock(DeviceExtension);

        return FALSE;
    }

    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerReleaseRemoveLock
//      Releases remove lock.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      None.
//
VOID TVTunerReleaseRemoveLock(
    IN  PTVTUNER_DEVICE_EXTENSION    DeviceExtension
    )
{
    LONG    count;

    count = InterlockedDecrement(&DeviceExtension->RemoveCount);

    ASSERT(count >= 0);

    if (count == 0)
    {
        KeSetEvent(&DeviceExtension->RemoveEvent, IO_NO_INCREMENT, FALSE);
    }

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerWaitForSafeRemove
//      Waits for all remove locks to be released
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      None.
//
//  Comment:
//      This routine should be called with no remove locks held 
//      by the calling thread
//
VOID TVTunerWaitForSafeRemove(
    IN  PTVTUNER_DEVICE_EXTENSION    DeviceExtension
    )
{
    DeviceExtension->PnpState = PnpStateRemoved;
    TVTunerReleaseRemoveLock(DeviceExtension);

    KeWaitForSingleObject(
        &DeviceExtension->RemoveEvent, 
        Executive, 
        KernelMode, 
        FALSE, 
        NULL
        );

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerStallQueues
//      Pauses all of the queues, and waits for them to get to paused state.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      none
//
VOID TVTunerStallQueues(
    IN  PTVTUNER_DEVICE_EXTENSION DeviceExtension
    )
{
    ULONG   index;

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++");

    // stall IRP processing
    TVTunerLockIo(&DeviceExtension->IoLock);

    TVTunerWaitForStopIo(&DeviceExtension->IoLock);

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--");

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerRestartQueues
//      Restarts all paused queues.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      none
//
VOID TVTunerRestartQueues(
    IN  PTVTUNER_DEVICE_EXTENSION   DeviceExtension
    )
{
    ULONG   index;

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++");

    TVTunerUnlockIo(&DeviceExtension->IoLock);

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--");

    return;
}
///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerFlushQueues
//      Flush oustanding IRPs for closed file object.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  FileObject
//              about to be closed file object
//
//  Return Value:
//      none
//
VOID TVTunerFlushQueues(
    IN  PTVTUNER_DEVICE_EXTENSION   DeviceExtension,
    IN  PFILE_OBJECT            FileObject
    )
{
    ULONG   index;

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++");

    TVTunerFlushPendingIo(&DeviceExtension->IoLock, FileObject);

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--");

    return;
}
