// iorw.c
//
// Requires DDK Only
// File created on 8/12/2008
//

#include "pch.h"
#include "..\intrface.h"
#ifdef TVTUNER_WMI_TRACE
#include "iorw.tmh"
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerReadDispatch
//      Handled incoming read requests
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our device
//
//      IN  Irp
//              The read IRP to handle
//
//  Return Value:
//      NT status code
//
NTSTATUS TVTunerReadDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    NTSTATUS                        status;
    PIO_STACK_LOCATION              irpStack;
    PVOID                           readBuffer;
    ULONG                           readLength;

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    status = TVTunerCheckIoLock(&deviceExtension->IoLock, Irp);
    if (!NT_SUCCESS(status) || (status == STATUS_PENDING))
    {
        TVTunerDebugPrint(DBG_IO, DBG_WARN, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

        return status;
    }

    // Get our IRP stack location
    irpStack = IoGetCurrentIrpStackLocation(Irp);

    // Get the read buffer length
    readLength = irpStack->Parameters.Read.Length;
    if (readLength == 0)
    {
        // just complete 0 length request
        status = STATUS_SUCCESS;

        Irp->IoStatus.Information = 0;
        Irp->IoStatus.Status = status;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);

        TVTunerDecrementIoCount(&deviceExtension->IoLock);

        TVTunerDebugPrint(DBG_IO, DBG_WARN, __FUNCTION__"--. IRP %p, STATUS %x", Irp, status);

        return status;
    }

    readBuffer = MmGetSystemAddressForMdlSafe(Irp->MdlAddress, NormalPagePriority);

    status = TVTunerPipe0Io(
                deviceExtension, 
                Irp, 
                &deviceExtension->InterfaceInformation->Pipes[0]
                );

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerWriteDispatch
//      Handled incoming write requests
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our device
//
//      IN  Irp
//              The write IRP to handle
//
//  Return Value:
//      NT status code
//
NTSTATUS TVTunerWriteDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    NTSTATUS                        status;
    PIO_STACK_LOCATION              irpStack;
    PVOID                           writeBuffer;
    ULONG                           writeLength;

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    status = TVTunerCheckIoLock(&deviceExtension->IoLock, Irp);
    if (!NT_SUCCESS(status) || (status == STATUS_PENDING))
    {
        TVTunerDebugPrint(DBG_IO, DBG_WARN, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

        return status;
    }

    // Get our IRP stack location
    irpStack = IoGetCurrentIrpStackLocation(Irp);

    // Get the write buffer length
    writeLength = irpStack->Parameters.Write.Length;
    if (writeLength == 0)
    {
        // just complete 0 length request
        status = STATUS_SUCCESS;

        Irp->IoStatus.Information = 0;
        Irp->IoStatus.Status = status;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);

        TVTunerDecrementIoCount(&deviceExtension->IoLock);

        TVTunerDebugPrint(DBG_IO, DBG_WARN, __FUNCTION__"--. IRP %p, STATUS %x", Irp, status);

        return status;
    }

    writeBuffer = MmGetSystemAddressForMdlSafe(Irp->MdlAddress, NormalPagePriority);

    status = TVTunerPipe1Io(
                deviceExtension, 
                Irp, 
                &deviceExtension->InterfaceInformation->Pipes[1]
                );

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerDeviceIoControlDispatch
//      Handled incoming IOCTL requests
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our device
//
//      IN  Irp
//              The IOCTL IRP to handle
//
//  Return Value:
//      NT status code
//
NTSTATUS TVTunerDeviceIoControlDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PIO_STACK_LOCATION              irpStack;
    NTSTATUS                        status;
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    PVOID                           inputBuffer;
    ULONG                           inputLength;
    PVOID                           outputBuffer;
    ULONG                           outputLength;

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    // Get our IRP stack location
    irpStack = IoGetCurrentIrpStackLocation(Irp);

    // Get the buffer lengths
    inputLength = irpStack->Parameters.DeviceIoControl.InputBufferLength;
    outputLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;

	TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__" Control code: %d", irpStack->Parameters.DeviceIoControl.IoControlCode);

    switch (irpStack->Parameters.DeviceIoControl.IoControlCode) 
    {
    case IOCTL_INTERNAL_USB_SUBMIT_URB:
        TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__": IOCTL_INTERNAL_USB_SUBMIT_URB");

        // put the control IRP in our control queue
        status = TVTunerQueueIrp(&deviceExtension->IOCTL_INTERNAL_USB_SUBMIT_URBQueue, Irp);
        break;

    case IOCTL_RESET_PIPE:
        TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__": IOCTL_RESET_PIPE");

        status = TVTunerCheckIoLock(&deviceExtension->IoLock, Irp);
        if (!NT_SUCCESS(status) || (status == STATUS_PENDING))
        {
            TVTunerDebugPrint(DBG_IO, DBG_WARN, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

            return status;
        }

        // buffered ioctl
        inputBuffer = Irp->AssociatedIrp.SystemBuffer;
        outputBuffer = Irp->AssociatedIrp.SystemBuffer;

        status = STATUS_NOT_IMPLEMENTED;
        Irp->IoStatus.Information = 0;
        Irp->IoStatus.Status = status;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);

        TVTunerDecrementIoCount(&deviceExtension->IoLock);
        break;

/*    case IOCTL_GET_CONFIG_DESCRIPTOR:
        TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__": IOCTL_GET_CONFIG_DESCRIPTOR");

        status = TvModuleCheckIoLock(&deviceExtension->IoLock, Irp);
        if (!NT_SUCCESS(status) || (status == STATUS_PENDING))
        {
            TVTunerDebugPrint(DBG_IO, DBG_WARN, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

            return status;
        }

        // buffered ioctl
        inputBuffer = Irp->AssociatedIrp.SystemBuffer;
        outputBuffer = Irp->AssociatedIrp.SystemBuffer;

        status = STATUS_NOT_IMPLEMENTED;
        Irp->IoStatus.Information = 0;
        Irp->IoStatus.Status = status;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);

        TVTunerDecrementIoCount(&deviceExtension->IoLock);
        break;

    case IOCTL_RESET_DEVICE:
        TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__": IOCTL_RESET_DEVICE");

        status = TVTunerCheckIoLock(&deviceExtension->IoLock, Irp);
        if (!NT_SUCCESS(status) || (status == STATUS_PENDING))
        {
            TVTunerDebugPrint(DBG_IO, DBG_WARN, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

            return status;
        }

        // buffered ioctl
        inputBuffer = Irp->AssociatedIrp.SystemBuffer;
        outputBuffer = Irp->AssociatedIrp.SystemBuffer;

        status = STATUS_NOT_IMPLEMENTED;
        Irp->IoStatus.Information = 0;
        Irp->IoStatus.Status = status;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);

        TVTunerDecrementIoCount(&deviceExtension->IoLock);
        break; */

    default:
        TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__" status == STATUS_INVALID_DEVICE_REQUEST");

        status = STATUS_INVALID_DEVICE_REQUEST;
        Irp->IoStatus.Status = status;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);
        break;
    }

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}
