// power.c
//
// Requires DDK Only
// File created on 8/12/2008
//

#include "pch.h"
#ifdef TVTUNER_WMI_TRACE
#include "power.tmh"
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerPowerDispatch 
//      Dispatch routine for IRPs of IRP_MJ_POWER type
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our driver
//
//      IN  Irp
//              The power IRP to handle
//
//  Return Value:
//      Status
//
NTSTATUS TVTunerPowerDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    NTSTATUS                        status;
    PIO_STACK_LOCATION              irpStack;

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    TVTunerDumpIrp(Irp);

    // Get our current IRP stack location
    irpStack = IoGetCurrentIrpStackLocation(Irp);

    // Get our device extension
    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    // check if device has been removed
    if (!TVTunerAcquireRemoveLock(deviceExtension))
    {
        status = STATUS_NO_SUCH_DEVICE;

        PoStartNextPowerIrp(Irp);
        Irp->IoStatus.Status = status;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);

        TVTunerDebugPrint(DBG_POWER, DBG_WARN, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

        return status;
    }

    // If our device has not been started, we just pass the 
    // power requests down the stack
    if (deviceExtension->PnpState == PnpStateNotStarted) 
    {
        PoStartNextPowerIrp(Irp);
        IoSkipCurrentIrpStackLocation(Irp);

        status = PoCallDriver(deviceExtension->LowerDeviceObject, Irp);

        // Release Remove Lock
        TVTunerReleaseRemoveLock(deviceExtension);

        TVTunerDebugPrint(DBG_POWER, DBG_WARN, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

        return status;
    }

    // Determine the power IRP type
    switch(irpStack->MinorFunction)
    {
    case IRP_MN_QUERY_POWER:

        if (irpStack->Parameters.Power.Type == SystemPowerState)
        {
            // Handle the system power IRP case
            status = TVTunerSystemPowerDispatch(deviceExtension, Irp);
        }
        else
        {
            // Handle the device power IRP case here

            // we should only get a query when going to a less or same powered state
            ASSERT(irpStack->Parameters.Power.State.DeviceState >= deviceExtension->DevicePowerState);

            // If the query is for our current state, just return SUCCESS
            if (irpStack->Parameters.Power.State.DeviceState == deviceExtension->DevicePowerState)
            {
                // Let the power manager know we can handle new device power IRPs
                PoStartNextPowerIrp(Irp);

                // forward the IRP down the stack
                IoSkipCurrentIrpStackLocation(Irp);
                Irp->IoStatus.Status = STATUS_SUCCESS;
                
                status = PoCallDriver(deviceExtension->LowerDeviceObject, Irp);

                // Adjust our power IRP count
                TVTunerReleaseRemoveLock(deviceExtension);
            }
            else
            {
                // Save off our device power IRP
                deviceExtension->DevicePowerIrp = Irp;

                // Mark the IRP pending
                IoMarkIrpPending(Irp);
                status = STATUS_PENDING;

                // handle device power IRP at PASSIVE_LEVEL
                if (KeGetCurrentIrql() < DISPATCH_LEVEL)
                {
                    TVTunerPowerDownPrepCallback(DeviceObject, deviceExtension);
                }
                else
                {
                    IoQueueWorkItem(
                        deviceExtension->PowerWorkItem,
                        TVTunerPowerDownPrepCallback,
                        DelayedWorkQueue,
                        deviceExtension
                        );
                }
            }
        }
        break;
    case IRP_MN_SET_POWER:

        if (irpStack->Parameters.Power.Type == SystemPowerState)
        {
            // Handle the system power IRP case
            status = TVTunerSystemPowerDispatch(deviceExtension, Irp);
        }
        else
        {
            // Handle the device power IRP case here

            // determine if we are powering up, powering down, or staying the same
            if (irpStack->Parameters.Power.State.DeviceState < deviceExtension->DevicePowerState)
            {
                // Powering up

                // Powering up needs to happen bottom to top for the device stack
                // so we send the IRP down with a completion routine, and handle
                // our power up tasks in the completion routine

                // Mark the IRP pending, as our completion 
                // routine won't complete it
                IoMarkIrpPending(Irp);
                status = STATUS_PENDING;

                // Prepare the next stack location
                IoCopyCurrentIrpStackLocationToNext(Irp);

                // Insert our completion routine
                IoSetCompletionRoutine(
                    Irp,
                    TVTunerDeviceSetPowerCompletionRoutine,
                    deviceExtension,
                    TRUE,
                    TRUE,
                    TRUE
                    );

                // send the IRP down the stack
                PoCallDriver(deviceExtension->LowerDeviceObject, Irp);
            }
            else if (irpStack->Parameters.Power.State.DeviceState > deviceExtension->DevicePowerState)
            {
                // Powering down

                // Powering down needs to occur top to bottom for the device stack, so
                // we can handle power down functionality right here

                //*****************************************************************
                //*****************************************************************
                // TODO:  Add device specific power down code here
                //*****************************************************************
                //*****************************************************************

                // Save off our device power IRP
                deviceExtension->DevicePowerIrp = Irp;

                // Mark the IRP pending
                IoMarkIrpPending(Irp);
                status = STATUS_PENDING;

                if (KeGetCurrentIrql() < DISPATCH_LEVEL)
                {
                    TVTunerPowerDownPrepCallback(DeviceObject, deviceExtension);
                }
                else
                {
                    IoQueueWorkItem(
                        deviceExtension->PowerWorkItem,
                        TVTunerPowerDownPrepCallback,
                        DelayedWorkQueue,
                        deviceExtension
                        );
                }
            }
            else
            {
                // Power stays the same

                // If our current device power state is PowerDeviceD0, then restart
                // our queues
                if (deviceExtension->DevicePowerState == PowerDeviceD0)
                {
                    // Save off our device power IRP
                    deviceExtension->DevicePowerIrp = Irp;

                    // Mark the IRP pending
                    IoMarkIrpPending(Irp);
                    status = STATUS_PENDING;

                    if (KeGetCurrentIrql() < DISPATCH_LEVEL)
                    {
                        TVTunerPowerD0PrepCallback(DeviceObject, deviceExtension);
                    }
                    else
                    {
                        IoQueueWorkItem(
                            deviceExtension->PowerWorkItem,
                            TVTunerPowerD0PrepCallback,
                            DelayedWorkQueue,
                            deviceExtension
                            );
                    }
                }
                else
                {
                    // Let the power manager know we can handle new
                    // device power IRPs
                    PoStartNextPowerIrp(Irp);

                    // forward the IRP down the stack
                    IoSkipCurrentIrpStackLocation(Irp);
                    Irp->IoStatus.Status = STATUS_SUCCESS;
                    
                    status = PoCallDriver(deviceExtension->LowerDeviceObject, Irp);

                    // Adjust our IRP count
                    TVTunerReleaseRemoveLock(deviceExtension);
                }
            }
        }

        break;
    case IRP_MN_WAIT_WAKE:
    case IRP_MN_POWER_SEQUENCE:
    default:
        // default case, just send the IRP down and let other drivers
        // in the stack handle it

        // Let the power manager know we can handle another
        // power IRP of this type
        PoStartNextPowerIrp(Irp);

        // send the IRP down the stack
        IoSkipCurrentIrpStackLocation(Irp);

        // Drivers must use PoCallDriver, rather than IoCallDriver,
        // to pass power IRPs. PoCallDriver allows the Power Manager
        // to ensure that power IRPs are properly synchronized throughout
        // the system.
        status = PoCallDriver(deviceExtension->LowerDeviceObject, Irp);

        // Adjust our IRP count
        TVTunerReleaseRemoveLock(deviceExtension);

        break;
    }

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);
    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerSystemPowerDispatch 
//      Dispatch routine for system IRP_MJ_POWER IRPs
//
//  Arguments:
//      IN  DeviceExtension
//              Our device extension
//
//      IN  Irp
//              The power IRP to handle
//
//  Return Value:
//      NT status code
//
NTSTATUS TVTunerSystemPowerDispatch(
    PTVTUNER_DEVICE_EXTENSION    DeviceExtension,
    IN  PIRP                        Irp
    )
{
    // In this case we have to request a device power IRP
    // before completing the system power IRP.  So, we send
    // the system power IRP down the stack with a completion
    // routine and request the device power IRP from that
    // completion routine.  We specify a power complete callback
    // when requesting the device power IRP, and we complete
    // the system power IRP in that power complete callback.

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    // Mark the IRP pending
    IoMarkIrpPending(Irp);

    // Prepare the next stack location
    IoCopyCurrentIrpStackLocationToNext(Irp);

    // Insert our completion routine
    IoSetCompletionRoutine(
        Irp,
        TVTunerSystemPowerCompletionRoutine,
        DeviceExtension,
        TRUE,
        TRUE,
        TRUE
        );

    // send the IRP down the stack
    PoCallDriver(DeviceExtension->LowerDeviceObject, Irp);

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"--. IRP %p", Irp);

    return STATUS_PENDING;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerSystemPowerCompletionRoutine
//      I/O completion routine for system power IRP requests.
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our device
//
//      IN  Irp
//              The system power IRP
//
//      IN  Context
//              Our device extension
//
//  Return Value:
//      returns STATUS_MORE_PROCESSING_REQUIRED if all goes well, 
//          indicating that we aren't done with the system power IRP
//      returns STATUS_SUCCESS if there is an error in order to let 
//          the system IRP go
//
NTSTATUS TVTunerSystemPowerCompletionRoutine(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp,
    IN  PVOID           Context
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    PIO_STACK_LOCATION              irpStack;
    POWER_STATE                     powerState;
    NTSTATUS                        status;

    // Get our device extension
    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)Context;

    status = Irp->IoStatus.Status;

    // Get the current stack location from the IRP
    irpStack = IoGetCurrentIrpStackLocation(Irp);

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++. IRP %p STATUS %x", Irp, status);

    if (NT_SUCCESS(status))
    {
        // Save off the system power IRP
        deviceExtension->SystemPowerIrp = Irp;

        // Get our device power state for the given system power state
        status = TVTunerGetDevicePowerState(
                    deviceExtension, 
                    irpStack->Parameters.Power.State.SystemState,
                    &powerState.DeviceState
                    );

        if (NT_SUCCESS(status))
        {
            // Request the device power IRP
            status = PoRequestPowerIrp(
                        deviceExtension->PhysicalDeviceObject,
                        irpStack->MinorFunction,  
                        powerState,
                        TVTunerDevicePowerCompleteCallback,
                        deviceExtension,
                        NULL
                        );
        }
    }

    if (!NT_SUCCESS(status))
    {
        ASSERT(irpStack->MinorFunction != IRP_MN_SET_POWER);

        // NULL our system IRP storage
        deviceExtension->SystemPowerIrp = NULL;

        // Let the power manager know we can handle new 
        // system power IRPs
        PoStartNextPowerIrp(Irp);

        // save error status
        Irp->IoStatus.Status = status;

        // Adjust our IRP count
        TVTunerReleaseRemoveLock(deviceExtension);

        TVTunerDebugPrint(DBG_POWER, DBG_WARN, __FUNCTION__"--. Irp %p STATUS %x", Irp, status);

        return STATUS_SUCCESS;
    }
    else
    {
        TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"--. Irp %p STATUS %x", Irp, status);

        // We aren't done with the system power IRP yet,
        // so we return more processing required here
        return STATUS_MORE_PROCESSING_REQUIRED;
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerDevicePowerCompleteCallback
//      Power complete callback routine called by the power manager when our 
//      device power IRP has been completely processed by the entire
//      device stack.  **This is not an I/O completion routine**
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our device
//
//      IN  MinorFunction
//              The minor function of the power IRP
//
//      IN  PowerState
//              The power state of the power IRP
//
//      IN  Context
//              Our device extension
//
//      IN  IoStatus
//              Status block of the power IRP
//
//  Return Value:
//      none
//
VOID TVTunerDevicePowerCompleteCallback(
    IN  PDEVICE_OBJECT      DeviceObject,
    IN  UCHAR               MinorFunction,
    IN  POWER_STATE         PowerState,
    IN  PVOID               Context,
    IN  PIO_STATUS_BLOCK    IoStatus
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    PIO_STACK_LOCATION              irpStack;

    // Get our device extension
    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)Context;

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++. IRP %p", deviceExtension->SystemPowerIrp);

    // Get the current stack location for our system power IRP
    irpStack = IoGetCurrentIrpStackLocation(deviceExtension->SystemPowerIrp);

    // If this is a QUERY power IRP, use the device power IRP's status
    // as the system power IRPs status
    if (MinorFunction == IRP_MN_QUERY_POWER)
    {
        deviceExtension->SystemPowerIrp->IoStatus.Status = IoStatus->Status;
    }
    else
    {
        ASSERT(MinorFunction == IRP_MN_SET_POWER);

        // do not fail set power irp
        deviceExtension->SystemPowerIrp->IoStatus.Status = STATUS_SUCCESS;

        // save the new system power state
        deviceExtension->SystemPowerState = irpStack->Parameters.Power.State.SystemState;
    }

    // Let the power manager know we can accept new system power IRPs
    PoStartNextPowerIrp(deviceExtension->SystemPowerIrp);

    // complete the system power IRP
    IoCompleteRequest(deviceExtension->SystemPowerIrp, IO_NO_INCREMENT);

    // set our pointer to NULL
    deviceExtension->SystemPowerIrp = NULL;

    // Adjust our IRP count
    TVTunerReleaseRemoveLock(deviceExtension);

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"--");

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerDeviceSetPowerCompletionRoutine
//      I/O completion routine for device power IRP requests to power up
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our device
//
//      IN  Irp
//              The system power IRP
//
//      IN  Context
//              Our device extension
//
//  Return Value:
//      returns STATUS_MORE_PROCESSING_REQUIRED
//
NTSTATUS TVTunerDeviceSetPowerCompletionRoutine(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp,
    IN  PVOID           Context
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    PIO_STACK_LOCATION              irpStack;
    NTSTATUS                        status;

    status = Irp->IoStatus.Status;

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++. IRP %p STATUS %x", Irp, status);

    // Get our device extension
    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)Context;

    if (NT_SUCCESS(status))
    {
        // Get our IRP stack location
        irpStack = IoGetCurrentIrpStackLocation(Irp);

        // Store off the device power IRP
        deviceExtension->DevicePowerIrp = Irp;

        // do the real work at PASSIVE_LEVEL
        if (KeGetCurrentIrql() < DISPATCH_LEVEL)
        {
            TVTunerPowerUpCallback(DeviceObject, deviceExtension);
        }
        else
        {
            // Queue the work item
            IoQueueWorkItem(
                deviceExtension->PowerWorkItem,
                TVTunerPowerUpCallback,
                DelayedWorkQueue,
                deviceExtension
                );
        }

        TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"--");

        // We aren't done with the device power IRP
        // so return more processing required
        return STATUS_MORE_PROCESSING_REQUIRED;
    }
    else
    {
        PoStartNextPowerIrp(Irp);

        // Adjust our IRP count
        TVTunerReleaseRemoveLock(deviceExtension);

        TVTunerDebugPrint(DBG_POWER, DBG_WARN, __FUNCTION__"--. STATUS %x", status);

        return STATUS_SUCCESS;
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerGetDevicePowerState
//      Uses the device capabilities structure, received in the IRP_MJ_PNP, 
//      IRP_MN_QUERY_CAPABILITIES IRP handler, to determine the System
//      to Device (S to D) power state mappings for a given system
//      power state
//
//  Arguments:
//      IN  DeviceExtension
//              Our device extension
//
//      IN  SystemState
//              System power state to provide device power state mapping for
//
//
//  Return Value:
//      returns the device power state corresponding to the system 
//      power state
//
NTSTATUS TVTunerGetDevicePowerState(
    IN  PTVTUNER_DEVICE_EXTENSION    DeviceExtension,
    IN  SYSTEM_POWER_STATE              SystemState,
    OUT PDEVICE_POWER_STATE             DeviceState
    )
{
    NTSTATUS            status;
    DEVICE_POWER_STATE  deviceState;

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++. %s", SystemPowerStateString(SystemState));

    status = STATUS_SUCCESS;

    if (SystemState == PowerSystemWorking)
    {
        *DeviceState = PowerDeviceD0;
    }
    else
    {
        *DeviceState = PowerDeviceD3;
    }

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++. %s STATUS %x", DevicePowerStateString(*DeviceState), status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerPowerDownPrepCallback
//      Callback used to handle power down tasks that need to either block
//      run at PASSIVE_LEVEL for some other reason
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our device
//
//      IN  Context
//              Our device extension
//
//  Return Value:
//      none
//
VOID TVTunerPowerDownPrepCallback(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PVOID           Context
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    PIRP                            irp;
    PIO_STACK_LOCATION              irpStack;

    // Get our device extension
    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)Context;

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++. IRP %p", deviceExtension->DevicePowerIrp);

    // get our device IRP
    irp = deviceExtension->DevicePowerIrp;

    // NULL our pointer
    deviceExtension->DevicePowerIrp = NULL;

    ASSERT(irp != NULL);

    irpStack = IoGetCurrentIrpStackLocation(irp);

    if (!deviceExtension->bPowerStop)
    {
        // stall all io processing
        deviceExtension->bPowerStop = TRUE;
        TVTunerStallQueues(deviceExtension);
    }

    if (irpStack->MinorFunction == IRP_MN_SET_POWER)
    {
        //*****************************************************************
        //*****************************************************************
        // TODO: power down your device
        //*****************************************************************
        //*****************************************************************

        // Let the power manager know our new device power state
        PoSetPowerState(
            deviceExtension->DeviceObject,
            irpStack->Parameters.Power.Type,
            irpStack->Parameters.Power.State
            );

        // Save our new power state
        deviceExtension->DevicePowerState = irpStack->Parameters.Power.State.DeviceState;
    }

    // Let the power manager know we can handle a new device power IRP
    PoStartNextPowerIrp(irp);

    // send the device power IRP down the stack
    IoSkipCurrentIrpStackLocation(irp);
    irp->IoStatus.Status = STATUS_SUCCESS;

    PoCallDriver(deviceExtension->LowerDeviceObject, irp);

    // Adjust our IRP count
    TVTunerReleaseRemoveLock(deviceExtension);

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"--");

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerPowerD0PrepCallback
//      Callback used to handle power down tasks that need to either block
//      run at PASSIVE_LEVEL for some other reason
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our device
//
//      IN  Context
//              Our device extension
//
//  Return Value:
//      none
//
VOID TVTunerPowerD0PrepCallback(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PVOID           Context
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    PIRP                    irp;
    PIO_STACK_LOCATION      irpStack;

    // Get our device extension
    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)Context;

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++. IRP %p", deviceExtension->DevicePowerIrp);

    //*****************************************************************
    //*****************************************************************
    // TODO:  Handle any unwinding of a previous query to power down here
    //*****************************************************************
    //*****************************************************************

    if (deviceExtension->bPowerStop)
    {
        // restart our queues
        deviceExtension->bPowerStop = FALSE;
        TVTunerRestartQueues(deviceExtension);
    }

    // get our device IRP
    irp = deviceExtension->DevicePowerIrp;

    // NULL our pointer
    deviceExtension->DevicePowerIrp = NULL;

    ASSERT(irp != NULL);

    // Let the power manager know we can handle new
    // device power IRPs
    PoStartNextPowerIrp(irp);

    // forward the IRP down the stack
    IoSkipCurrentIrpStackLocation(irp);
    irp->IoStatus.Status = STATUS_SUCCESS;
    
    PoCallDriver(deviceExtension->LowerDeviceObject, irp);

    // Adjust our IRP count
    TVTunerReleaseRemoveLock(deviceExtension);

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"--");

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerPowerUpCallback
//      Callback used to handle power down tasks that need to either block
//      run at PASSIVE_LEVEL for some other reason
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our device
//
//      IN  Context
//              Our device extension
//
//  Return Value:
//      none
//
VOID TVTunerPowerUpCallback(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PVOID           Context
    )
{
    PTVTUNER_DEVICE_EXTENSION    deviceExtension;
    PIRP                            irp;
    PIO_STACK_LOCATION              irpStack;

    // Get our device extension
    deviceExtension = (PTVTUNER_DEVICE_EXTENSION)Context;

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++. IRP %p", deviceExtension->DevicePowerIrp);

    // Get our device power IRP
    irp = deviceExtension->DevicePowerIrp;

    ASSERT(irp != NULL);

    // NULL our IRP pointer
    deviceExtension->DevicePowerIrp = NULL;

    //*****************************************************************
    //*****************************************************************
    // TODO:  Add device specific power up code here
    //*****************************************************************
    //*****************************************************************

    // Get our IRP stack location
    irpStack = IoGetCurrentIrpStackLocation(irp);

    // Let the power manager know our new device power state
    PoSetPowerState(
        deviceExtension->DeviceObject,
        irpStack->Parameters.Power.Type,
        irpStack->Parameters.Power.State
        );

    // Save our new power state
    deviceExtension->DevicePowerState = irpStack->Parameters.Power.State.DeviceState;

    if (deviceExtension->bPowerStop)
    {
        // restart our queues
        deviceExtension->bPowerStop = FALSE;
        TVTunerRestartQueues(deviceExtension);
    }

    // Let the power manager know we can handle new
    // device power IRPs
    PoStartNextPowerIrp(irp);

    // Complete the IRP
    IoCompleteRequest(irp, IO_NO_INCREMENT);

    // Adjust our IRP count
    TVTunerReleaseRemoveLock(deviceExtension);

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"--");

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerRequestDevicePowerIrpCallback
//      Callback used to signal completion of power handling
//
//  Arguments:
//      IN  DeviceObject
//              Device object for our device
//
//      IN  MinorFunction
//              The minor function of the power IRP
//
//      IN  PowerState
//              The power state of the power IRP
//
//      IN  Context
//              Our device extension
//
//      IN  IoStatus
//              Status block of the power IRP
//
//  Return Value:
//      none
//
VOID TVTunerRequestDevicePowerIrpCallback(
    IN  PDEVICE_OBJECT      DeviceObject,
    IN  UCHAR               MinorFunction,
    IN  POWER_STATE         PowerState,
    IN  PVOID               Context,
    IN  PIO_STATUS_BLOCK    IoStatus
    )
{
    PTVTUNER_POWER_IRP_CONTEXT context;

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++");

    context = (PTVTUNER_POWER_IRP_CONTEXT)Context;

    // copy status in our context block
    context->Status = IoStatus->Status;

    // signal completion event
    KeSetEvent(context->Event, IO_NO_INCREMENT, FALSE);

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"--");

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerRequestDevicePowerIrp
//      This routine requests IRP_MN_SET_POWER and wait for its
//      completion
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  DeviceState
//              New device power state
//
//  Return Value:
//      NT Status code
//
NTSTATUS TVTunerRequestDevicePowerIrp(
    IN  PTVTUNER_DEVICE_EXTENSION        DeviceExtension,
    IN  DEVICE_POWER_STATE             DeviceState
    )
{
    NTSTATUS                    status;
    POWER_STATE                 powerState;
    TVTUNER_POWER_IRP_CONTEXT   powerContext;
    KEVENT                      event;

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"++");

    // We must be at IRQL less than DISPATCH_LEVEL
    // in order to block on KeWaitForSingleObject
    ASSERT(KeGetCurrentIrql() < DISPATCH_LEVEL);

    KeInitializeEvent(&event, NotificationEvent, FALSE);
    powerContext.Event = &event;
    powerState.DeviceState = DeviceState;

    status = PoRequestPowerIrp(
                DeviceExtension->PhysicalDeviceObject,
                IRP_MN_SET_POWER,
                powerState,
                TVTunerRequestDevicePowerIrpCallback,
                &powerContext,
                NULL
                );

    if (status == STATUS_PENDING)
    {
        KeWaitForSingleObject(
            &event, 
            Executive,
            KernelMode,
            FALSE,
            NULL
            );

        status = powerContext.Status;
    }

    TVTunerDebugPrint(DBG_POWER, DBG_TRACE, __FUNCTION__"--. STATUS %x", status);

    return status;
}
