// usb.c
//
// Requires DDK Only
// File created on 8/12/2008
//

#include "pch.h"
#ifdef TVTUNER_WMI_TRACE
#include "usb.tmh"
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerSubmitUrbSynch
//      method to synchronously submit an urb to the bus driver.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  Urb
//              URB to submit
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerSubmitUrbSynch(
    IN  PTVTUNER_DEVICE_EXTENSION   DeviceExtension,
    IN  PURB                        Urb
    )
{
    NTSTATUS            status;
    PIRP                irp;
    IO_STATUS_BLOCK     ioStatus;
    KEVENT              event;
    PIO_STACK_LOCATION  irpStack;

    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"++");

    KeInitializeEvent(&event, NotificationEvent, FALSE);

    irp = IoBuildDeviceIoControlRequest(
            IOCTL_INTERNAL_USB_SUBMIT_URB,
            DeviceExtension->LowerDeviceObject,
            NULL,
            0,
            NULL,
            0,
            TRUE,
            &event,
            &ioStatus
            );


    if (irp != NULL)
    {
	    TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"++. IRP %p", irp);


        TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" IoBuildDeviceIoControlRequest is done...");

        irpStack = IoGetNextIrpStackLocation(irp);
        irpStack->Parameters.Others.Argument1 = Urb;

        TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" prepeare to call IoCallDriver function...");

        status = IoCallDriver(DeviceExtension->LowerDeviceObject, irp);
        if (status == STATUS_PENDING)
        {
            TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" retreivied status is STATUS_PENDING...");

            KeWaitForSingleObject(
                &event,
                Executive,
                KernelMode,
                FALSE,
                NULL
                );
            
            status = ioStatus.Status;    
        }
    }
    else
    {
        TVTunerDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__" IoBuildDeviceIoControlRequest is failed. irp == NULL");

        status = STATUS_INSUFFICIENT_RESOURCES;
    }

    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerResetPipe
//      clears endpoint stall condition.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  PipeHandle
//              Handle to a pipe, which needs to be reset
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerResetPipe(
    IN  PTVTUNER_DEVICE_EXTENSION   DeviceExtension,
    IN  USBD_PIPE_HANDLE                    PipeHandle
    )
{
    NTSTATUS                    status;
    struct _URB_PIPE_REQUEST    urb;

    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"++");

    urb.Hdr.Length = (USHORT)sizeof(struct _URB_PIPE_REQUEST);
    urb.Hdr.Function = URB_FUNCTION_RESET_PIPE;
    urb.PipeHandle = PipeHandle;

    status = TVTunerSubmitUrbSynch(DeviceExtension, (PURB)&urb);

    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerAbortPipe
//      Aborts all the outstanding requests on the pipe.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  PipeHandle
//              Handle to a pipe, which needs to be aborted
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerAbortPipe(
    IN  PTVTUNER_DEVICE_EXTENSION   DeviceExtension,
    IN  USBD_PIPE_HANDLE                    PipeHandle
    )
{
    NTSTATUS                    status;
    struct _URB_PIPE_REQUEST    urb;

    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"++");

    urb.Hdr.Length = (USHORT)sizeof(struct _URB_PIPE_REQUEST);
    urb.Hdr.Function = URB_FUNCTION_ABORT_PIPE;
    urb.PipeHandle = PipeHandle;

    status = TVTunerSubmitUrbSynch(DeviceExtension, (PURB)&urb);

    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerResetDevice
//      Resets the port associated with the PDO.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerResetDevice(
    IN  PTVTUNER_DEVICE_EXTENSION   DeviceExtension
    )
{
    NTSTATUS            status;
    PIRP                irp;
    KEVENT              event;
    IO_STATUS_BLOCK     ioStatus;

    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"++");

    KeInitializeEvent(&event, NotificationEvent, FALSE);

    irp = IoBuildDeviceIoControlRequest(
            IOCTL_INTERNAL_USB_RESET_PORT,
            DeviceExtension->LowerDeviceObject,
            NULL,
            0,
            NULL,
            0,
            TRUE,
            &event,
            &ioStatus
            );

    if(irp != NULL) 
    {
        status = IoCallDriver(DeviceExtension->LowerDeviceObject, irp);
        if (status == STATUS_PENDING)
        {
            KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
            status = ioStatus.Status;
        }
    }
    else
    {
        status = STATUS_INSUFFICIENT_RESOURCES;
    }

    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerSelectAlternateInterface
//      This routine selects an alternate interface in the same
//      configuration.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  AlternateSetting
//              interface number
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerSelectAlternateInterface(
    IN  PTVTUNER_DEVICE_EXTENSION   DeviceExtension,
    IN  ULONG                               AlternateSetting
    )
{
    NTSTATUS                    status;
    PUSB_INTERFACE_DESCRIPTOR   interfaceDescriptor;
    PURB                        urb;
    ULONG                       urbSize;
    PUSBD_INTERFACE_INFORMATION interfaceInfo;
    ULONG                       index;

    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"++");

    // find new interface descriptor in the configuration descriptor
    interfaceDescriptor = USBD_ParseConfigurationDescriptorEx(
                                DeviceExtension->ConfigDescriptor,
                                DeviceExtension->ConfigDescriptor,
                                0,
                                AlternateSetting,
                                -1,
                                -1,
                                -1
                                );


    if (interfaceDescriptor != NULL)
    {
        // allocate select interface URB
        urbSize = GET_SELECT_INTERFACE_REQUEST_SIZE(
                        interfaceDescriptor->bNumEndpoints
                        );

        urb = (PURB)ExAllocatePoolWithTag(
                        NonPagedPool,
                        urbSize,
                        TVTUNER_POOL_TAG
                        );

        if (urb != NULL)
        {
            RtlZeroMemory(urb, urbSize);

            UsbBuildSelectInterfaceRequest(
                urb,
                (USHORT)urbSize,
                DeviceExtension->ConfigHandle,
                0,
                (UCHAR)AlternateSetting
                );

            interfaceInfo = &urb->UrbSelectInterface.Interface;
            interfaceInfo->Length = 
                GET_USBD_INTERFACE_SIZE(interfaceDescriptor->bNumEndpoints);

            //*****************************************************************
            //*****************************************************************
            // TODO: initialize MaximumTransferSize with appropriate values
            //*****************************************************************
            //*****************************************************************

			// USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE
		    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE %d", USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE);

            for (index = 0; index < interfaceDescriptor->bNumEndpoints; ++index)
            {
                interfaceInfo->Pipes[index].MaximumTransferSize = USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE; 
            }

            status = TVTunerSubmitUrbSynch(DeviceExtension, urb);
            if (NT_SUCCESS(status))
            {
                // save new interface information in our device extension
                ExFreePool(DeviceExtension->InterfaceInformation);

                DeviceExtension->InterfaceInformation = 
                    (PUSBD_INTERFACE_INFORMATION)ExAllocatePoolWithTag(
                                                    NonPagedPool, 
                                                    interfaceInfo->Length, 
                                                    TVTUNER_POOL_TAG
                                                    );

                if (DeviceExtension->InterfaceInformation != NULL)
                {
                    RtlCopyMemory(
                        DeviceExtension->InterfaceInformation, 
                        interfaceInfo, 
                        interfaceInfo->Length
                        );
                }
                else
                {
                    status = STATUS_INSUFFICIENT_RESOURCES;
                }
            }

            ExFreePool(urb);
        }
        else
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
        }
    }
    else
    {
        status = STATUS_DEVICE_CONFIGURATION_ERROR;
    }


    TVTunerDebugPrint(DBG_IO, DBG_INFO, __FUNCTION__"--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerPipe0Io
//      performs read from or writes to bulk pipe.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  Irp
//              Io request
//
//      IN  PipeInformation
//              Bulk pipe
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerPipe0Io(
    IN  PTVTUNER_DEVICE_EXTENSION   DeviceExtension,
    IN  PIRP                                Irp,
    IN  PUSBD_PIPE_INFORMATION              PipeInformation
    )
{
    PTVTUNER_IO_CONTEXT ioContext;
    NTSTATUS                    status;
    PUCHAR                      virtualAddress;
    ULONG                       totalLength;
    ULONG                       transferLength;
    PMDL                        mdl;
    PURB                        urb;
    PIO_STACK_LOCATION          irpStack;

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    // we will use do{}while(FALSE); structure
    // for resources cleanup
    status = STATUS_SUCCESS;
    ioContext = NULL;
    mdl = NULL;

    do
    {
        Irp->IoStatus.Information = 0;

        // allocate io context to pass to completion routine
        ioContext = (PTVTUNER_IO_CONTEXT)ExAllocatePoolWithTag(
                                                    NonPagedPool, 
                                                    sizeof(TVTUNER_IO_CONTEXT), 
                                                    TVTUNER_POOL_TAG
                                                    );

        if (ioContext == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        if (Irp->MdlAddress != NULL)
        {
            totalLength = MmGetMdlByteCount(Irp->MdlAddress); 
        }
        else
        {
            totalLength = 0;
        }

        if (totalLength == 0)
        {
            status = STATUS_SUCCESS;
            break;
        }

        virtualAddress = (PUCHAR)MmGetMdlVirtualAddress(Irp->MdlAddress);

        if (totalLength > USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE)
        {
            transferLength = USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE;
        }
        else
        {
            transferLength = totalLength;
        }

        mdl = IoAllocateMdl(virtualAddress, totalLength, FALSE, FALSE, NULL);
        if (mdl == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        IoBuildPartialMdl(Irp->MdlAddress, mdl, virtualAddress, transferLength);

        urb = (PURB)ExAllocatePoolWithTag(
                        NonPagedPool, 
                        sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER), 
                        TVTUNER_POOL_TAG
                        );

        if (urb == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        UsbBuildInterruptOrBulkTransferRequest(
            urb,
            sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER),
            PipeInformation->PipeHandle,
            NULL,
            mdl,
            transferLength,
            USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK,
            NULL
            );

        // save io parameters in our io context
        ioContext->DeviceExtension = DeviceExtension;
        ioContext->NextVirtualAddress = virtualAddress + transferLength;
        ioContext->RemainingLength = totalLength - transferLength;
        ioContext->TotalTransferCount = 0;
        ioContext->Mdl = mdl;
        ioContext->Urb = urb;

        irpStack = IoGetNextIrpStackLocation(Irp);
        irpStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
        irpStack->Parameters.Others.Argument1 = urb;
        irpStack->Parameters.DeviceIoControl.IoControlCode = IOCTL_INTERNAL_USB_SUBMIT_URB;

        IoSetCompletionRoutine(
            Irp, 
            TVTunerPipe0IoCompletionRoutine,
            ioContext,
            TRUE,
            TRUE,
            TRUE
            );

        IoMarkIrpPending(Irp);

        IoCallDriver(DeviceExtension->LowerDeviceObject, Irp);
        status = STATUS_PENDING;
        ioContext = NULL;
        mdl = NULL;
    }
    while (FALSE);

    if (ioContext != NULL)
    {
        ExFreePool(ioContext);
    }

    if (mdl != NULL)
    {
        IoFreeMdl(mdl);
    }

    if (status != STATUS_PENDING)
    {
        Irp->IoStatus.Status = status;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);

        TVTunerDecrementIoCount(&DeviceExtension->IoLock);
    }

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerPipe0IoCompletionRoutine
//      Completion routine for bulk transfer.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the create IRP
//
//      IN  Context
//              our io context
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerPipe0IoCompletionRoutine(
    IN  PDEVICE_OBJECT DeviceObject,
    IN  PIRP           Irp,
    IN  PVOID          Context
    )
{
    PTVTUNER_IO_CONTEXT     ioContext;
    NTSTATUS                        status;
    ULONG                           transferLength;
    PIO_STACK_LOCATION              irpStack;

    ioContext = (PTVTUNER_IO_CONTEXT)Context;
    status = Irp->IoStatus.Status;

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++. IRP %p STATUS %x", Irp, status);

    if (NT_SUCCESS(status))
    {
        ioContext->TotalTransferCount += ioContext->Urb->UrbBulkOrInterruptTransfer.TransferBufferLength;

        if (ioContext->RemainingLength)
        {
            if (ioContext->RemainingLength > USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE)
            {
                transferLength = USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE;
            }
            else
            {
                transferLength = ioContext->RemainingLength;
            }

            MmPrepareMdlForReuse(ioContext->Mdl);
            IoBuildPartialMdl(Irp->MdlAddress, ioContext->Mdl, ioContext->NextVirtualAddress, transferLength);

            ioContext->Urb->UrbBulkOrInterruptTransfer.TransferBufferLength = transferLength;

            ioContext->RemainingLength -= transferLength;
            ioContext->NextVirtualAddress += transferLength;

            irpStack = IoGetNextIrpStackLocation(Irp);
            irpStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
            irpStack->Parameters.Others.Argument1 = ioContext->Urb;
            irpStack->Parameters.DeviceIoControl.IoControlCode = IOCTL_INTERNAL_USB_SUBMIT_URB;

            IoSetCompletionRoutine(
                Irp, 
                TVTunerPipe0IoCompletionRoutine,
                ioContext,
                TRUE,
                TRUE,
                TRUE
                );

            IoCallDriver(ioContext->DeviceExtension->LowerDeviceObject, Irp);

            TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--. Next Stage");

            return STATUS_MORE_PROCESSING_REQUIRED;
        }
        else
        {
            Irp->IoStatus.Information = ioContext->TotalTransferCount;
        }
    }

    // free the resources
    ExFreePool(ioContext->Urb);
    IoFreeMdl(ioContext->Mdl);
    TVTunerDecrementIoCount(&ioContext->DeviceExtension->IoLock);
    ExFreePool(ioContext);

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--. Done");

    return STATUS_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerPipe1Io
//      performs read from or writes to bulk pipe.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  Irp
//              Io request
//
//      IN  PipeInformation
//              Bulk pipe
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerPipe1Io(
    IN  PTVTUNER_DEVICE_EXTENSION   DeviceExtension,
    IN  PIRP                                Irp,
    IN  PUSBD_PIPE_INFORMATION              PipeInformation
    )
{
    PTVTUNER_IO_CONTEXT ioContext;
    NTSTATUS                    status;
    PUCHAR                      virtualAddress;
    ULONG                       totalLength;
    ULONG                       transferLength;
    PMDL                        mdl;
    PURB                        urb;
    PIO_STACK_LOCATION          irpStack;

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    // we will use do{}while(FALSE); structure
    // for resources cleanup
    status = STATUS_SUCCESS;
    ioContext = NULL;
    mdl = NULL;

    do
    {
        Irp->IoStatus.Information = 0;

        // allocate io context to pass to completion routine
        ioContext = (PTVTUNER_IO_CONTEXT)ExAllocatePoolWithTag(
                                                    NonPagedPool, 
                                                    sizeof(TVTUNER_IO_CONTEXT), 
                                                    TVTUNER_POOL_TAG
                                                    );

        if (ioContext == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        if (Irp->MdlAddress != NULL)
        {
            totalLength = MmGetMdlByteCount(Irp->MdlAddress); 
        }
        else
        {
            totalLength = 0;
        }

        if (totalLength == 0)
        {
            status = STATUS_SUCCESS;
            break;
        }

        virtualAddress = (PUCHAR)MmGetMdlVirtualAddress(Irp->MdlAddress);

        if (totalLength > USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE)
        {
            transferLength = USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE;
        }
        else
        {
            transferLength = totalLength;
        }

        mdl = IoAllocateMdl(virtualAddress, totalLength, FALSE, FALSE, NULL);
        if (mdl == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        IoBuildPartialMdl(Irp->MdlAddress, mdl, virtualAddress, transferLength);

        urb = (PURB)ExAllocatePoolWithTag(
                        NonPagedPool, 
                        sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER), 
                        TVTUNER_POOL_TAG
                        );

        if (urb == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        UsbBuildInterruptOrBulkTransferRequest(
            urb,
            sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER),
            PipeInformation->PipeHandle,
            NULL,
            mdl,
            transferLength,
            USBD_TRANSFER_DIRECTION_OUT,
            NULL
            );

        // save io parameters in our io context
        ioContext->DeviceExtension = DeviceExtension;
        ioContext->NextVirtualAddress = virtualAddress + transferLength;
        ioContext->RemainingLength = totalLength - transferLength;
        ioContext->TotalTransferCount = 0;
        ioContext->Mdl = mdl;
        ioContext->Urb = urb;

        irpStack = IoGetNextIrpStackLocation(Irp);
        irpStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
        irpStack->Parameters.Others.Argument1 = urb;
        irpStack->Parameters.DeviceIoControl.IoControlCode = IOCTL_INTERNAL_USB_SUBMIT_URB;

        IoSetCompletionRoutine(
            Irp, 
            TVTunerPipe1IoCompletionRoutine,
            ioContext,
            TRUE,
            TRUE,
            TRUE
            );

        IoMarkIrpPending(Irp);

        IoCallDriver(DeviceExtension->LowerDeviceObject, Irp);
        status = STATUS_PENDING;
        ioContext = NULL;
        mdl = NULL;
    }
    while (FALSE);

    if (ioContext != NULL)
    {
        ExFreePool(ioContext);
    }

    if (mdl != NULL)
    {
        IoFreeMdl(mdl);
    }

    if (status != STATUS_PENDING)
    {
        Irp->IoStatus.Status = status;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);

        TVTunerDecrementIoCount(&DeviceExtension->IoLock);
    }

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  TVTunerPipe1IoCompletionRoutine
//      Completion routine for bulk transfer.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the create IRP
//
//      IN  Context
//              our io context
//
//  Return Value:
//      NT status code.
//
NTSTATUS TVTunerPipe1IoCompletionRoutine(
    IN  PDEVICE_OBJECT DeviceObject,
    IN  PIRP           Irp,
    IN  PVOID          Context
    )
{
    PTVTUNER_IO_CONTEXT     ioContext;
    NTSTATUS                        status;
    ULONG                           transferLength;
    PIO_STACK_LOCATION              irpStack;

    ioContext = (PTVTUNER_IO_CONTEXT)Context;
    status = Irp->IoStatus.Status;

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++. IRP %p STATUS %x", Irp, status);

    if (NT_SUCCESS(status))
    {
        ioContext->TotalTransferCount += ioContext->Urb->UrbBulkOrInterruptTransfer.TransferBufferLength;

        if (ioContext->RemainingLength)
        {
            if (ioContext->RemainingLength > USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE)
            {
                transferLength = USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE;
            }
            else
            {
                transferLength = ioContext->RemainingLength;
            }

            MmPrepareMdlForReuse(ioContext->Mdl);
            IoBuildPartialMdl(Irp->MdlAddress, ioContext->Mdl, ioContext->NextVirtualAddress, transferLength);

            ioContext->Urb->UrbBulkOrInterruptTransfer.TransferBufferLength = transferLength;

            ioContext->RemainingLength -= transferLength;
            ioContext->NextVirtualAddress += transferLength;

            irpStack = IoGetNextIrpStackLocation(Irp);
            irpStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
            irpStack->Parameters.Others.Argument1 = ioContext->Urb;
            irpStack->Parameters.DeviceIoControl.IoControlCode = IOCTL_INTERNAL_USB_SUBMIT_URB;

            IoSetCompletionRoutine(
                Irp, 
                TVTunerPipe1IoCompletionRoutine,
                ioContext,
                TRUE,
                TRUE,
                TRUE
                );

            IoCallDriver(ioContext->DeviceExtension->LowerDeviceObject, Irp);

            TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--. Next Stage");

            return STATUS_MORE_PROCESSING_REQUIRED;
        }
        else
        {
            Irp->IoStatus.Information = ioContext->TotalTransferCount;
        }
    }

    // free the resources
    ExFreePool(ioContext->Urb);
    IoFreeMdl(ioContext->Mdl);
    TVTunerDecrementIoCount(&ioContext->DeviceExtension->IoLock);
    ExFreePool(ioContext);

    TVTunerDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--. Done");

    return STATUS_SUCCESS;
}
